/*
 * Copyright 1994  	Doug Anson, danson@lgc.com & David Holland, davidh@use.com
 *
 * File:   aperture.c
 * Author: Doug Anson (danson@lgc.com)
 * Date:   1/31/94
 * 
 * Modified: David Holland (davidh@use.com)
 * Date:     2/23/94
 * Changes: Changed general name of driver
 * 			Debugging flags
 * 			Removed a few extraneous erorr messages.
 * 		
 *
 * Purpose: This file is the master device file for the Solaris 2.1 x86
 *          framebuffer aperture driver. Portions of this driver are taken
 *          from mmap.c 1.4 copyright 93/06/03 Sun  Microsystems, Inc.
 *          (Sample memory mapped driver for Solaris 2.0/Sun OS 5.0)
 *
 *      	This driver allows accelerated SVGA graphics cards that are
 *			attached vi 32-bit paths (VESA/EISA/MC) and that have 
 *			linear framebuffer capabilities to have their framebuffer
 *			memory mapped into the user process space.
 *
 * Disclamer:
 *
 * This code is based largely upon the sample device drivers provided
 * by Sun Microsystems Inc.
 * 
 * Original copyright notice:
 * @(#)mmap.c     1.3 copyright 93/06/03 Sun Microsystems, Inc.
 *
 * Our copyright notice:
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Doug Anson, and David Holland be used in
 * advertising or publicity pertaining to distribution of the software 
 * Doug Anson, and David Holland make no * representations about the 
 * suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * Original disclamer from Sun Microsystems, Inc.:
 * This is a package of sample device drivers for Solaris 2.x (SunOS
 * 5.x).  You may use, modify and distribute these drivers and/or binaries
 * derived from them. However please note that:
 * 
 * These examples are provided with no warranties of any kind, including
 * without limitation accuracy and usefulness, and Sun expressly disclaims
 * all implied warranties of merchantability, fitness for a particular
 * purpose and non-infringement. In no event shall Sun be liable for any
 * damages, including without limitation, direct, special, indirect, or
 * consequential damages arising out of, or relating to, use of these
 * examples by customer or any third party. Sun is under no obligation to
 * provide support to customer for this software.
 * 
 * Our disclaimer:
 * DOUG ANSON, AND DAVID HOLLAND DISCLAIMS ALL WARRIENTS WITH REGARD TO THIS 
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY, AND FITNESS, 
 * IN NO EVENT SHALL DOUG ANSON, OR DAVID HOLLAND BE LIABLE FOR ANY SPECIAL, 
 * INDIRECT, OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING FROM 
 * USAGE OF THIS SOFTWARE.
 */

/*
 * linear framebuffer aperture driver
 * 
 * $Id
 */

#include <sys/types.h>
#include <sys/param.h>
#include <sys/errno.h>
#include <sys/file.h>
#include <sys/conf.h>
#include <sys/uio.h>
#include <sys/map.h>
#include <sys/debug.h>
#include <sys/modctl.h>
#include <sys/kmem.h>
#include <sys/cmn_err.h>
#include <sys/open.h>
#include <sys/stat.h>
#include <sys/ddi.h>
#include <sys/sunddi.h>

#include "aperture.h"

/* opaque handle top of state structs */
static	void *state_head = NULL;

/* 
 * Solaris DDI/DKI driver entry points
 */
#if defined(__STDC__)
static	int	aperture_getinfo(dev_info_t *, ddi_info_cmd_t, void *,void **);
static	int	aperture_identify(dev_info_t *dip);
static  int aperture_probe(dev_info_t *dip);
static	int	aperture_attach(dev_info_t *, ddi_attach_cmd_t);
static	int	aperture_open(dev_t *, int, int, cred_t *);
static	int	aperture_close(dev_t, int, int, cred_t *);
static	int	aperture_detach(dev_info_t *, ddi_detach_cmd_t);
static	int	aperture_mmap(dev_t, off_t, int);
#else
static  int	aperture_getinfo();
static  int	aperture_identify();
static  int	aperture_probe();
static  int	aperture_attach();
static  int	aperture_open();
static  int	aperture_close();
static  int	aperture_detach();
static  int	aperture_mmap();
#endif /* __STDC__ */

/*
 * Solaris DDI/DKI module linkeage structure declarations
 */
static struct cb_ops	aperture_cb_ops = 
{
	aperture_open,			/* XXopen routine						*/
	aperture_close,			/* XXclose routine						*/
	nodev,					/* XXstrategy routine					*/
	nodev,					/* XXprint routine	    				*/
	nodev,					/* XXdump routine	    				*/
	nodev,					/* XXread routine	    				*/
	nodev,					/* XXwrite routine 						*/
	nodev,					/* XXioctl routine 						*/
	nodev,					/* XXdevmap routine						*/
	aperture_mmap,			/* XXmmap routine						*/
	ddi_segmap,				/* XXsegmap routine						*/
	nochpoll,				/* XXchpoll routine						*/
	ddi_prop_op,			/* XXprop_op routine					*/	
	(struct streamtab *)0,	/* not a STREAMS driver				 	*/
	D_NEW | D_MP,			/* safe for multi-thread/multi-processor*/
};

static struct dev_ops aperture_ops = 
{
	DEVO_REV,				/* DEVO_REV indicated by manual		*/
	0,						/* device reference count	    	*/
	aperture_getinfo,		/* devo_getinfo routine				*/
	aperture_identify,		/* devo_identiry routine			*/
	aperture_probe,			/* device probe for non-self-id 	*/
	aperture_attach,		/* devo_attach routine				*/
	aperture_detach,		/* devo_detach routine				*/
	nodev,					/* device reset routine				*/
	&aperture_cb_ops,		/* cb_ops structure pointer			*/
	(struct bus_ops *)0,	/* bus operations					*/
};

extern	struct	mod_ops mod_driverops;
static	struct modldrv modldrv = 
{
	&mod_driverops,					/* mod_ops structure pointer 	*/
	DEV_BANNER_STRING,				/* device banner string			*/
	&aperture_ops,					/* dev_ops structure pointer 	*/
};

static	struct modlinkage modlinkage = 
{
	MODREV_1,			/* as indicated indicated by manual 	 		*/
	(void *)&modldrv,	/* module driver structure pointer				*/
	NULL,				/* termination of list of linkage structures 	*/
};

/*
 * Function: _init()
 * Purpose:  Solaris DDI/DKI _init(9E) entry point
 * Author:   Doug Anson
 * Calls:     
 * Log:		DMA -- Origional Version
 * 		DWH - Removed extra cmn_err() calls 	2/23/94
 * 		DWH - Changed debugging 		2/23/94
 */
#if defined(__STDC__)
int _init(void)
#else
int _init()
#endif /* __STDC__ */
{
	register int	error;

#if APERTURE_DEBUG >= 3

        /* enterance announcement */
        cmn_err(CE_CONT,"%s: entering _init(aperture)\n", DEV_IDENT_STRING);

#endif

	if ((error = ddi_soft_state_init(&state_head, sizeof (Mmap), 1)) != 0)
	{
		/* failed to get state structure */
		return (error);
	}

	if ((error = mod_install(&modlinkage)) != 0)
		ddi_soft_state_fini(&state_head);

	return (error);
}

/*
 * Function: _info()
 * Purpose:  Solaris DDI/DKI _info(9E) entry point
 * Author:   Doug Anson
 * Calls:     
 * Modified: David Holland (davidh@use.com)
 * Log:      DMA -- Origional Version
 * 			 DWH - Changed debugging 	2/23/94
 */
#if defined(__STDC__)
int _info(struct modinfo *modinfop)
#else
int _info(modinfop)
struct modinfo *modinfop;
#endif /* __STDC__ */
{

#if APERTURE_DEBUG >= 3

	/* entrance announcement */
	cmn_err(CE_CONT,"%s: entering _info(aperture)\n", DEV_IDENT_STRING);

#endif /* APERTURE_DEBUG >= 3 */

	return (mod_info(&modlinkage, modinfop));
}

/*
 * Function: _fini()
 * Purpose:  Solaris DDI/DKI _fini(9E) entry point
 * Author:   Doug Anson
 * Calls:     
 * Modified: David Holland (davidh@use.com)
 * Log:      DMA -- Origional Version
 * 			 DWH - Changed debugging
 */
#if defined(__STDC__)
int _fini(void)
#else
int _fini()
#endif /* __STDC__ */
{
	int status;

#if APERTURE_DEBUG >= 3

/* entrance announcement */
cmn_err(CE_CONT,"%s: entering _fini(aperture)\n", DEV_IDENT_STRING);

#endif /* APERTURE_DEBUG >= 3*/

	if ((status = mod_remove(&modlinkage)) != 0)
		return (status);

	ddi_soft_state_fini(&state_head);

	return (status);
}

/*
 * Function: aperture_getinfo()
 * Purpose:  This function provides the getinfo(9E) functionality 
 * Author:   Doug Anson
 * Calls:     
 * Modified: David Holland (davidh@use.com)
 * Log:      DMA -- Origional Version
 * 			 DWH - Changed function name/variables/debugging	2/23/94
 */
#if defined(__STDC__)
static int aperture_getinfo(dev_info_t *dip, ddi_info_cmd_t infocmd, 
						  void *arg, void **result)
#else
static int aperture_getinfo(dip,infocmd,arg,result)
dev_info_t     *dip;
ddi_info_cmd_t infocmd;
void           *arg;
void           **result;
#endif /* __STDC__ */
{
	register int 	error;
	register Mmap	*aperture_p = NULL;

#if APERTURE_DEBUG >= 3

        /* enterance announcement */
        cmn_err(CE_CONT,"%s: entering aperture_getinfo()\n", DEV_IDENT_STRING);

#endif

	switch (infocmd) 
	{
	    case DDI_INFO_DEVT2DEVINFO:
			aperture_p = (Mmap *)ddi_get_soft_state(state_head, getminor((dev_t)arg));
			if (aperture_p == NULL) 
			{
				/* could not get state structure */
				*result = NULL;
				error = DDI_FAILURE;
			} 
			else 
			{
				/* got state structure */
				*result = aperture_p->dip;
				error = DDI_SUCCESS;
			}
			break;
		case DDI_INFO_DEVT2INSTANCE:
			*result = (void *)getminor((dev_t)arg);
			error = DDI_SUCCESS;
			break;
		default:
			*result = NULL;
			error = DDI_FAILURE;
			break;
	}
	return (error);
}

/*
 * Function: aperture_identify()
 * Purpose:  This function implements the identify(9E) functionality
 * Author:   Doug Anson
 * Calls:     
 * Modified: David Holland (davidh@use.com)
 * Log:      DMA -- Origional Version
 * 			 DWH - Changed function name/debugging/variables	2/23/94
 */
#if defined(__STDC__)
static int aperture_identify(dev_info_t *dip)
#else
static int aperture_identify(dip)
dev_info_t *dip;
#endif /* __STDC__ */
{
#if APERTURE_DEBUG >= 3

        /* entrance announcement */
        cmn_err(CE_CONT,"%s: entering aperture_identify()\n", DEV_IDENT_STRING);

#endif 

	/* check the driver name and respond */
	if (strcmp(ddi_get_name(dip), APERTURE_NAME) == 0)
		return (DDI_IDENTIFIED); 		/* device driver is identified */
	else
		return (DDI_NOT_IDENTIFIED);	/* device driver not identified */
}

/*
 * Function: aperture_probe()
 * Purpose:  This function implements the probe(9E) functionality.
 *		     Currently this function returns a "dont care" result
 *			 as nothing is probed.
 * Author:   Doug Anson
 * Calls:     
 * Modfied:	 David Holland (davidh@use.com)
 * Log:      DMA -- Origional Version
 * 			 DWH - changed function name/variables/debugging	2/23/94
 */
#if defined(__STDC__)
static int aperture_probe(dev_info_t *dip)
#else
static int aperture_probe(dip)
dev_info_t *dip;
#endif /* __STDC__ */
{

#if APERTURE_DEBUG >= 3

        /* enterance announcement */
        cmn_err(CE_CONT,"%s: entering aperture_probe()\n", DEV_IDENT_STRING);

#endif

	/* return success as we really aren't probing for anything */
	return (DDI_PROBE_SUCCESS);
}

/*
 * Function: aperture_attach()
 * Purpose:  This function implements the attach(9E) functionality
 * Author:   Doug Anson
 * Calls:     
 * Modified: David Holland (davidh@use.com)
 * Log:      DMA -- Origional Version
 * 			 DWH - changed function name/variables/debugging	2/23/94
 */
#if defined(__STDC__)
static int aperture_attach(dev_info_t *dip, ddi_attach_cmd_t cmd)
#else
static int aperture_attach(dip,cmd)
dev_info_t       *dip;
ddi_attach_cmd_t cmd;
#endif /* __STDC__ */
{
	register Mmap  *aperture_p = NULL;
	int			   buf[AP_MAX];
	int			   length;
	int			   instance;

#if APERTURE_DEBUG >= 3

        /* entrance announcement */
        cmn_err(CE_CONT,"%s: entering aperture_attach()\n", DEV_IDENT_STRING);

#endif

	/* only continue if doing a DDI_ATTACH */
	if (cmd == DDI_ATTACH)
	{
		/* get and initialize our state structure */
		instance = ddi_get_instance(dip);
		if (ddi_soft_state_zalloc(state_head, instance) != DDI_SUCCESS)
		{
			/* could not allocate state structure */
			return (DDI_FAILURE);
		}

		/* get the state structure */
		aperture_p = (Mmap *)ddi_get_soft_state(state_head, instance);

		if (aperture_p == NULL)
		{
			/* failed to get the soft state */
			return (DDI_FAILURE);
		}

		/* initialize the mutex driver (ATOMIC) */
		mutex_init(&aperture_p->map_lock, "apperture mmap lock", MUTEX_DRIVER, 0);

		/* ENTER the critical region */
		mutex_enter(&aperture_p->map_lock);
		
		/* get the "reg" property from the .conf file */
		length = sizeof(buf);
		if (ddi_getlongprop_buf(DDI_DEV_T_ANY, dip, DDI_PROP_DONTPASS, "reg", (caddr_t)buf, &length) != DDI_PROP_SUCCESS)
		{
			/* could not get the framebuffer properties */
			return (DDI_FAILURE);
		}

		/* remember our register configuration */
        aperture_p->regnum  = (u_int)buf[AP_REGNUM];
        aperture_p->regbase = (caddr_t)buf[AP_ADDR];
        aperture_p->regsize = (off_t)buf[AP_SIZE];

		/* initialize the mmap state info */
		aperture_p->reglength = (off_t)AP_MEM_CHUNK;
		aperture_p->kaddr	 = (caddr_t)0;
        aperture_p->off     = (off_t)0;
        aperture_p->mapaddr = AP_MEM_FAILURE;

		/* initialize internal accounting */
        aperture_p->dip     = dip;

		/* EXIT the critical region */
		mutex_exit(&aperture_p->map_lock);

		/* create the device node */
		/* DDI_PSEUDO, CLONE_DEV */
		if (ddi_create_minor_node(dip, ddi_get_name(dip), S_IFCHR, instance, NULL, (int)NULL) == DDI_FAILURE) 
		{
			/* failed to make the device leaf or in error state */
			mutex_destroy(&aperture_p->map_lock);
			ddi_soft_state_free(state_head, instance);
			return (DDI_FAILURE);
		}

		/* device driver attached successfully */
		ddi_report_dev(aperture_p->dip);
		return (DDI_SUCCESS);
	}
	else
	{
		/* command not DDI_ATTACH -- return failure */
		return (DDI_FAILURE);
	}
}

/*
 * Function: aperture_detach()
 * Purpose:  This function implements the detach(9E) functionality
 * Author:   Doug Anson
 * Calls:     
 * Modified: David Holland (davidh@use.com)
 * Log:      DMA -- Origional Version
 * 			 DWH - changed function name/variables/debugging	2/23/94
 */
#if defined(__STDC__)
static int aperture_detach(dev_info_t *dip, ddi_detach_cmd_t cmd)
#else
static int aperture_detach(dip,cmd)
dev_info_t       *dip;
ddi_detach_cmd_t cmd;
#endif /* __STDC__ */
{
	register Mmap  	*aperture_p = NULL;	
	int		       	instance;

#if APERTURE_DEBUG >= 3

	/* entrance announcement */
	cmn_err(CE_CONT,"%s: entering aperture_detach()\n", DEV_IDENT_STRING);

#endif

	/* check the parameter list */
	if (cmd != DDI_DETACH)
	{
		return (DDI_FAILURE);
	}

	/* get the instance info and state structure */
	instance = ddi_get_instance(dip);
	aperture_p = (Mmap *)ddi_get_soft_state(state_head, instance);
	if (aperture_p == NULL)
	{
		/* failed to get the soft state */
		return (DDI_FAILURE);
	}

	/* clean up the driver */
	mutex_destroy(&aperture_p->map_lock);
	ddi_remove_minor_node(dip, NULL);
	ddi_soft_state_free(state_head, instance);

    /* device driver detached successfully */
	return (DDI_SUCCESS);
}

/*
 * Function: aperture_open()
 * Purpose:  This function is called in response to the open(2) system call
 * Author:   Doug Anson
 * Calls:     
 * Modified: David Holland (davidh@use.com)
 * Log:      DMA -- Origional Version
 * 			 DWH - Changed function name/variables/debugging
 */
#if defined(__STDC__)
static	int aperture_open(dev_t *dev, int openflags, int otyp, cred_t *credp)
#else
static	int aperture_open(dev, openflags, otyp, credp)
dev_t  *dev;
int    openflags;
int    otyp;
cred_t *credp;
#endif /* __STDC__ */
{
	register Mmap	*aperture_p = NULL;

#if APERTURE_DEBUG >= 3

	/* entrance announcement */
	cmn_err(CE_CONT,"%s: entering aperture_open()\n", DEV_IDENT_STRING);

#endif

	/* get the state structure */
	aperture_p = (Mmap *)ddi_get_soft_state(state_head, getminor(*dev));
	if (aperture_p == NULL)
	{
		/* failed to get the soft state */
		return (ENXIO);
	}

	/* check the parameter list for correctness */
	if (otyp != OTYP_CHR)
	{
		/* invalid parameter list */
		return (EINVAL);
	}

	return (0);
}

/*
 * Function: aperture_close()
 * Purpose:  This function is called after the last process that has 
 * 		     the device open calls close(2)
 * Author:   Doug Anson
 * Calls:     
 * Modified: David Holland (davidh@use.com)
 * Log:      DMA -- Origional Version
 * 			 DWH - changed function name/variables/debugging	2/23/94
 */
#if defined(__STDC__)
static	int aperture_close(dev_t dev, int closeflags, int otyp, cred_t *credp)
#else
static	int aperture_close(dev, closeflags, otyp, credp)
dev_t  dev;
int    closeflags;
int    otyp;
cred_t *credp;
#endif /* __STDC__ */
{
	register Mmap	*aperture_p = NULL;

#if APERTURE_DEBUG >= 3

        /* enterance announcement */
        cmn_err(CE_CONT,"%s: entering aperture_close()\n", DEV_IDENT_STRING);

#endif 

	/* get the state structure */
    aperture_p = (Mmap *)ddi_get_soft_state(state_head, getminor(dev));
    if (aperture_p == NULL)
	{
		/* failed to get the soft state */
        return (ENXIO);
	}

	/* check the paramter list for correctness */
    if (otyp != OTYP_CHR)
	{
		/* invalid parameter */
        return (EINVAL);
	}

	/* do nothing but return successful */
    return (0);
}

/*
 * Function: aperture_mmap()
 * Purpose:  This function gets called when the user process tries to mmap 
 *			 the framebuffer device using mmap(2).
 * Author:   Doug Anson
 * Calls:     
 * Modified: David Holland (davidh@use.com)
 * Log:      DMA -- Origional Version
 * 			 DWH - changed function name/variables/debugging	2/23/94
 */
#if defined(__STDC__)
static int aperture_mmap(dev_t dev, off_t off, int prot)
#else
static int aperture_mmap(dev, off, prot)
dev_t dev;
off_t off;
int   prot;
#endif /* __STDC__ */
{
	register Mmap	*aperture_p = NULL;

#if APERTURE_DEBUG >= 3

        /* enterance announcement */
        cmn_err(CE_CONT,"%s: entering aperture_mmap()\n", DEV_IDENT_STRING);

#endif

	/* get the state structure */
	aperture_p = (Mmap *)ddi_get_soft_state(state_head, getminor(dev));
	if (aperture_p == NULL)
	{
		/* failed to get the soft state */
		return (AP_MEM_FAILURE);
	}

	/* check that we dont mmap out of bounds */
	if (off > aperture_p->regsize)
	{
		/* offset is out of range for register length */
		return (AP_MEM_FAILURE);
	}

	/* check that READ/WRITE only priv only */
	if (!((prot & PROT_READ) || (prot & PROT_WRITE)))
	{
		/* invalid prot parameter in mmap(2) call */
        return (AP_MEM_FAILURE);
	} 

	/* ENTER the critical region */
	mutex_enter(&aperture_p->map_lock);

	/* map the registers  CRITICAL REGION ASSIGNMENT */

	aperture_p->off = off;
    if (ddi_map_regs(aperture_p->dip, aperture_p->regnum, (caddr_t *)&(aperture_p->kaddr), 
		aperture_p->off, aperture_p->reglength) == DDI_FAILURE)
	{

#if APERTURE_DEBUG >= 2

		/* report error */
		cmn_err(CE_CONT, "%s: ERROR: mmap failed (map regs). kaddr=0x%x off=0x%x\n", 
			DEV_IDENT_STRING,aperture_p->kaddr,aperture_p->off);

#endif

		aperture_p->mapaddr = (off_t)AP_MEM_FAILURE;
	}
	else
	{
		/* fill in our state structure  CRITICAL REGION ASSIGNMENT */
	   	aperture_p->mapaddr  = (off_t)hat_getkpfnum((caddr_t)aperture_p->kaddr); 
		if ((u_int)aperture_p->mapaddr == (u_int)AP_MEM_FAILURE)
		{

#if APERTURE_DEBUG >= 2

			/* report error */
			cmn_err(CE_CONT, "%s: ERROR: mmap failed (kpf). kaddr=0x%x off=0x%x\n", 
				DEV_IDENT_STRING,aperture_p->kaddr,aperture_p->off);
#endif 

		}

	    /* now unmap the region  CRITICAL REGION ASSIGNMENT */

       	 ddi_unmap_regs(aperture_p->dip, aperture_p->regnum, (caddr_t *)&(aperture_p->kaddr), 
			aperture_p->off, aperture_p->reglength);

#if APERTURE_DEBUG >= 2
  		/* report register mmaping status */
		if (aperture_p->mapaddr != AP_MEM_FAILURE)
		{
			/* mmap successful */
   			cmn_err(CE_CONT,"%s: mmap [off=0x%x,length=0x%x) successful\n", 
				DEV_IDENT_STRING,aperture_p->off,aperture_p->regsize);

    		cmn_err(CE_CONT,"%s: configuration:\n",DEV_IDENT_STRING);
    		cmn_err(CE_CONT,"    REG(%d): regbase=0x%x regsize=0x%x\n", 
				aperture_p->regnum, aperture_p->regbase, aperture_p->regsize);

    		cmn_err(CE_CONT,"    kaddr=0x%x mmap off=0x%x mmap addr=0x%x\n", 
				aperture_p->kaddr, aperture_p->off, aperture_p->mapaddr);
		}
		else
		{
			/* error in mmap routine */
			cmn_err(CE_CONT,"%s: mmap not established due to errors", DEV_IDENT_STRING);
		}
#endif

	}

    /* EXIT the critical region */
    mutex_exit(&aperture_p->map_lock);

	/* return the page frame number (PFN) */

#if APERTURE_DEBUG >= 3

	/* report exit - determining where the long delay is... */
	cmn_err(CE_CONT, "%s: _mmap routine exiting\n", DEV_IDENT_STRING);

#endif
	
	return ((int)(aperture_p->mapaddr));
}
