/*
 * Copyright 1994  	Doug Anson, danson@lgc.com & David Holland, davidh@use.com
 *
 * File:   aperture.h
 * Author: Doug Anson
 * Date:   1/31/94
 *
 * Modified: David Holland (davidh@use.com)
 * Log:		 Modified for new name		2/23/9
 *
 * Purpose: This header is the master header for the Solaris 2.1 x86
 *          framebuffer mmap driver. Portions of this driver are taken
 *          from mmapio.h 1.4 copyright 93/06/03 Sun  Microsystems, Inc.
 *
 * Disclamer:
 *
 * This code is based largely upon the sample device drivers provided
 * by Sun Microsystems Inc.
 * 
 * Original copyright notice:
 * @(#)mmapio.h 1.4 copyright 93/06/03 Sun Microsystems, Inc.
 *
 * Our copyright notice:
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Doug Anson, and David Holland be used in
 * advertising or publicity pertaining to distribution of the software 
 * Doug Anson, and David Holland make no * representations about the 
 * suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * Original disclamer from Sun Microsystems, Inc.:
 * This is a package of sample device drivers for Solaris 2.x (SunOS
 * 5.x).  You may use, modify and distribute these drivers and/or binaries
 * derived from them. However please note that:
 * 
 * These examples are provided with no warranties of any kind, including
 * without limitation accuracy and usefulness, and Sun expressly disclaims
 * all implied warranties of merchantability, fitness for a particular
 * purpose and non-infringement. In no event shall Sun be liable for any
 * damages, including without limitation, direct, special, indirect, or
 * consequential damages arising out of, or relating to, use of these
 * examples by customer or any third party. Sun is under no obligation to
 * provide support to customer for this software.
 * 
 * Our disclaimer:
 * DOUG ANSON, AND DAVID HOLLAND DISCLAIMS ALL WARRIENTS WITH REGARD TO THIS 
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY, AND FITNESS, 
 * IN NO EVENT SHALL DOUG ANSON, OR DAVID HOLLAND BE LIABLE FOR ANY SPECIAL, 
 * INDIRECT, OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING FROM 
 * USAGE OF THIS SOFTWARE.
 */

/*
 * linear framebuffer aperture driver header file
 * 
 * $Id
 */

#ifndef APERTURE_H
#define APERTURE_H

/*
 * Framebuffer identification definitions 
 */
#define APERTURE_NAME 		"aperture"
#define DEV_BANNER_STRING	"SVGA Framebuffer Driver v0.99"
#define DEV_IDENT_STRING	APERTURE_NAME	

/*
 * driver.conf(4) "reg" property definition for aperture driver: 
 * FORMAT:	reg=AP_REGNUM,AP_ADDR,AP_SIZE
 */
#define AP_REGNUM	0		/* register number is 1st number in "reg" prop */
#define AP_ADDR		1		/* fb base address is 2nd number in "reg" prop */
#define AP_SIZE		2		/* fb length value is 3rd number in "reg" prop */
#define AP_MAX		3		/* number of values in "reg" property		   */

/*
 * How big a chunk of register we map with each ddi_map_regs(9F)
 */
#define AP_MEM_CHUNK			sizeof(int)

/*
 * Error value for mmap(2) failure status 
 */
#define AP_MEM_FAILURE	-1

/*
 * Main memory map structure for the Framebuffer mmap driver
 */
struct mmap 
{
	u_int			 regnum;		/* registers number			   	   */
	volatile caddr_t regbase;		/* base address of device to mmap  */
	off_t		     regsize;		/* length of region to mmap(2)	   */
	off_t			 reglength;		/* map length for ddi_map_regs(9F) */
	volatile caddr_t kaddr;			/* kernel address of mapped device */
	off_t			 off;			/* offset from mmap(2) call		   */
	volatile off_t   mapaddr;		/* base address of mmap(2) region  */
	kmutex_t		 map_lock;		/* device map mutex lock		   */
	dev_info_t	     *dip;			/* device information pointer	   */
};
typedef struct mmap Mmap;

#endif /* APERTURE_H */
